@@ -15,6 +15,8 @@
 #include "udb.h"
 struct nsd;
 struct nsdst;
+struct ixfr_store;
+struct collect_rrs;
 
 #define DIFF_PART_XXFR ('X'<<24 | 'X'<<16 | 'F'<<8 | 'R')
 #define DIFF_PART_XFRF ('X'<<24 | 'F'<<16 | 'R'<<8 | 'F')
@@ -57,14 +59,16 @@ int diff_read_str(FILE* in, char* buf, size_t len);
 void delete_zone_rrs(namedb_type* db, zone_type* zone);
 /* delete an RR */
 int delete_RR(namedb_type* db, const dname_type* dname,
-	uint16_t type, uint16_t klass,
+	uint16_t type, uint16_t klass, uint32_t ttl,
 	buffer_type* packet, size_t rdatalen, zone_type *zone,
-	region_type* temp_region, int* softfail);
+	region_type* temp_region, int* softfail,
+	struct ixfr_store* ixfr_store);
 /* add an RR */
 int add_RR(namedb_type* db, const dname_type* dname,
 	uint16_t type, uint16_t klass, uint32_t ttl,
 	buffer_type* packet, size_t rdatalen, zone_type *zone,
-	int* softfail);
+	int* softfail, struct ixfr_store* ixfr_store,
+	struct collect_rrs* collect_rrs);
 
 /* apply the xfr file identified by xfrfilenr to zone */
 int apply_ixfr_for_zone(struct nsd* nsd, zone_type* zone, FILE* in,
