@@ -18,6 +18,7 @@
 #include <errno.h>
 #include <event2/event.h>
 #include <event2/http.h>
+#include <ctype.h>
 
 #include "nsd.h"
 #include "xfrd.h"
@@ -95,6 +96,7 @@ void daemon_metrics_close(struct daemon_metrics* metrics)
 
 	if (metrics->http_server) {
 		evhttp_free(metrics->http_server);
+		metrics->http_server = NULL;
 	}
 }
 
@@ -291,6 +293,10 @@ daemon_metrics_attach(struct daemon_metrics* metrics, struct xfrd_state* xfrd)
 	metrics->xfrd = xfrd;
 
 	metrics->http_server = evhttp_new(xfrd->event_base);
+	if(!metrics->http_server) {
+		log_msg(LOG_ERR, "metrics: out of memory in evhttp_new");
+		return;
+	}
 	for(p = metrics->accept_list; p; p = p->next) {
 		fd = p->accept_fd;
 		if (evhttp_accept_socket(metrics->http_server, fd)) {
@@ -343,6 +349,22 @@ metrics_http_callback(struct evhttp_request *req, void *p)
 }
 
 #ifdef BIND8_STATS
+/** Change disallowed characters, '.' ':' to underscores '_'. */
+static void
+change_string_underscores(char* prefix)
+{
+	/* Prometheus does not want '.' in the metric names. But the zone
+	 * statistics could have then in their name.
+	 * Also ':' is not allowed. This routine enforeces that it
+	 * has letters,digits,underscores. */
+	char* s = prefix;
+	while(*s) {
+		if(!isalnum((unsigned char)*s) && *s != '_')
+			*s = '_';
+		s++;
+	}
+}
+
 /** print long number*/
 static int
 print_longnum(struct evbuffer *buf, char* desc, uint64_t x)
@@ -381,13 +403,14 @@ print_stat_block(struct evbuffer *buf, struct nsdst* st,
 	char prefix[512] = {0};
 	if (name) {
 		snprintf(prefix, sizeof(prefix), "nsd_zonestats_%s_", name);
+		change_string_underscores(prefix);
 	} else {
-		sprintf(prefix, "nsd_");
+		snprintf(prefix, sizeof(prefix), "nsd_");
 	}
 
 	/* nsd_queries_by_type_total */
 	print_metric_help_and_type(buf, prefix, "queries_by_type_total",
-	                           "Total number of queries recieved by type.",
+	                           "Total number of queries received by type.",
 	                           "counter");
 	for(i=0; i<= 255; i++) {
 		if(metrics_inhibit_zero && st->qtype[i] == 0 &&
@@ -399,7 +422,7 @@ print_stat_block(struct evbuffer *buf, struct nsdst* st,
 
 	/* nsd_queries_by_class_total */
 	print_metric_help_and_type(buf, prefix, "queries_by_class_total",
-	                           "Total number of queries recieved by class.",
+	                           "Total number of queries received by class.",
 	                           "counter");
 	for(i=0; i<4; i++) {
 		if(metrics_inhibit_zero && st->qclass[i] == 0 && i != CLASS_IN)
@@ -410,7 +433,7 @@ print_stat_block(struct evbuffer *buf, struct nsdst* st,
 
 	/* nsd_queries_by_opcode_total */
 	print_metric_help_and_type(buf, prefix, "queries_by_opcode_total",
-	                           "Total number of queries recieved by opcode.",
+	                           "Total number of queries received by opcode.",
 	                           "counter");
 	for(i=0; i<6; i++) {
 		if(metrics_inhibit_zero && st->opcode[i] == 0 && i != OPCODE_QUERY)
@@ -421,7 +444,7 @@ print_stat_block(struct evbuffer *buf, struct nsdst* st,
 
 	/* nsd_queries_by_rcode_total */
 	print_metric_help_and_type(buf, prefix, "queries_by_rcode_total",
-	                           "Total number of queries recieved by rcode.",
+	                           "Total number of queries received by rcode.",
 	                           "counter");
 	for(i=0; i<17; i++) {
 		if(metrics_inhibit_zero && st->rcode[i] == 0 &&
@@ -433,20 +456,20 @@ print_stat_block(struct evbuffer *buf, struct nsdst* st,
 
 	/* nsd_queries_by_transport_total */
 	print_metric_help_and_type(buf, prefix, "queries_by_transport_total",
-		"Total number of queries recieved by transport.",
+		"Total number of queries received by transport.",
 		"counter");
 	evbuffer_add_printf(buf, "%squeries_by_transport_total{transport=\"udp\"} %lu\n", prefix, (unsigned long)st->qudp);
 	evbuffer_add_printf(buf, "%squeries_by_transport_total{transport=\"udp6\"} %lu\n", prefix, (unsigned long)st->qudp6);
 
 	/* nsd_queries_with_edns_total */
 	print_metric_help_and_type(buf, prefix, "queries_with_edns_total",
-		"Total number of queries recieved with EDNS OPT.",
+		"Total number of queries received with EDNS OPT.",
 		"counter");
 	evbuffer_add_printf(buf, "%squeries_with_edns_total %lu\n", prefix, (unsigned long)st->edns);
 
 	/* nsd_queries_with_edns_failed_total */
 	print_metric_help_and_type(buf, prefix, "queries_with_edns_failed_total",
-		"Total number of queries recieved with EDNS OPT where EDNS parsing failed.",
+		"Total number of queries received with EDNS OPT where EDNS parsing failed.",
 		"counter");
 	evbuffer_add_printf(buf, "%squeries_with_edns_failed_total %lu\n", prefix, (unsigned long)st->ednserr);
 
@@ -504,10 +527,11 @@ metrics_zonestat_print_one(struct evbuffer *buf, char *name,
 {
 	char prefix[512] = {0};
 	snprintf(prefix, sizeof(prefix), "nsd_zonestats_%s_", name);
+	change_string_underscores(prefix);
 
 	print_metric_help_and_type(buf, prefix, "queries_total",
-		"Total number of queries recieved.", "counter");
-	evbuffer_add_printf(buf, "nsd_zonestats_%s_queries_total %lu\n", name,
+		"Total number of queries received.", "counter");
+	evbuffer_add_printf(buf, "%squeries_total %lu\n", prefix,
 		(unsigned long)(zst->qudp + zst->qudp6 + zst->ctcp +
 			zst->ctcp6 + zst->ctls + zst->ctls6));
 	print_stat_block(buf, zst, name);
@@ -524,7 +548,7 @@ metrics_print_stats(struct evbuffer *buf, xfrd_state_type *xfrd,
 
 	/* nsd_queries_total */
 	print_metric_help_and_type(buf, "nsd_", "queries_total",
-	                           "Total number of queries recieved.", "counter");
+	                           "Total number of queries received.", "counter");
 	/*per CPU and total*/
 	for(i=0; i<xfrd->nsd->child_count; i++) {
 		evbuffer_add_printf(buf, "nsd_queries_total{server=\"%d\"} %lu\n",
