@@ -44,60 +44,6 @@
 #include "nsec3.h"
 #include "zone.h"
 
-/*
- * Compares two rdata arrays.
- *
- * Returns:
- *
- *	zero if they are equal
- *	non-zero if not
- *
- */
-static int
-zrdatacmp(uint16_t type, const union rdata_atom *rdatas, size_t rdata_count, rr_type *b)
-{
-	assert(rdatas);
-	assert(b);
-
-	/* One is shorter than another */
-	if (rdata_count != b->rdata_count)
-		return 1;
-
-	/* Compare element by element */
-	for (size_t i = 0; i < rdata_count; ++i) {
-		if (rdata_atom_is_domain(type, i)) {
-			if (rdata_atom_domain(rdatas[i])
-			    != rdata_atom_domain(b->rdatas[i]))
-			{
-				return 1;
-			}
-		} else if(rdata_atom_is_literal_domain(type, i)) {
-			if (rdata_atom_size(rdatas[i])
-			    != rdata_atom_size(b->rdatas[i]))
-				return 1;
-			if (!dname_equal_nocase(rdata_atom_data(rdatas[i]),
-				   rdata_atom_data(b->rdatas[i]),
-				   rdata_atom_size(rdatas[i])))
-				return 1;
-		} else {
-			if (rdata_atom_size(rdatas[i])
-			    != rdata_atom_size(b->rdatas[i]))
-			{
-				return 1;
-			}
-			if (memcmp(rdata_atom_data(rdatas[i]),
-				   rdata_atom_data(b->rdatas[i]),
-				   rdata_atom_size(rdatas[i])) != 0)
-			{
-				return 1;
-			}
-		}
-	}
-
-	/* Otherwise they are equal */
-	return 0;
-}
-
 /*
  * Find rrset type for any zone
  */
@@ -162,13 +108,99 @@ has_soa(domain_type* domain)
 struct zonec_state {
 	struct namedb *database;
 	struct domain_table *domains;
-	struct region *rr_region;
 	struct zone *zone;
-	struct domain *domain;
 	size_t errors;
 	size_t records;
+
+	struct collect_rrs c;
 };
 
+static void zonec_commit_rrset(zone_parser_t *parser, struct zonec_state *state)
+{
+	struct rrset *rrset;
+	int priority = parser->options.secondary ? ZONE_WARNING : ZONE_ERROR;
+
+	if(!state->c.domain || state->c.rr_count == 0)
+		return;
+	if (!state->c.rrset) {
+		rrset = region_alloc(state->database->region, sizeof(*rrset)
+#ifdef PACKED_STRUCTS
+			+ sizeof(rr_type*) * state->c.rr_count /* Add space for RRs. */
+#endif
+			);
+		rrset->zone = state->zone;
+		rrset->rr_count = state->c.rr_count;
+#ifndef PACKED_STRUCTS
+		rrset->rrs = region_alloc(state->database->region,
+				sizeof(rr_type*) * state->c.rr_count);
+#endif
+		memcpy(rrset->rrs, state->c.rrs, state->c.rr_count * sizeof(rr_type*));
+		switch (state->c.type) {
+			case TYPE_CNAME:
+				if (!domain_find_non_cname_rrset(state->c.domain, state->zone))
+					break;
+				zone_log(parser, priority, "CNAME and other data at the same name");
+				break;
+			case TYPE_RRSIG:
+			case TYPE_NXT:
+			case TYPE_SIG:
+			case TYPE_NSEC:
+			case TYPE_NSEC3:
+				break;
+			default:
+				if (!domain_find_rrset(state->c.domain, state->zone, TYPE_CNAME))
+					break;
+				zone_log(parser, priority, "CNAME and other data at the same name");
+				break;
+		}
+		/* Add it */
+		domain_add_rrset(state->c.domain, rrset);
+	} else {
+#ifndef PACKED_STRUCTS
+		struct rr **rrs;
+#else
+		struct rrset *rrset_orig;
+#endif
+		/* Add it... */
+		rrset = state->c.rrset;
+#ifndef PACKED_STRUCTS
+		rrs = rrset->rrs;
+		rrset->rrs = region_alloc_array(
+			state->database->region, rrset->rr_count + state->c.rr_count, sizeof(*rrs));
+		memcpy(rrset->rrs, rrs, rrset->rr_count * sizeof(*rrs));
+		region_recycle(state->database->region, rrs, rrset->rr_count * sizeof(*rrs));
+#else
+		rrset_orig = rrset;
+		rrset = region_alloc(state->database->region,
+			sizeof(rrset_type) +
+			(rrset_orig->rr_count+state->c.rr_count)*sizeof(rr_type*));
+		memcpy(rrset, rrset_orig,
+			sizeof(rrset_type) +
+			rrset_orig->rr_count*sizeof(rr_type*));
+		if(state->c.rrset_prev)
+			state->c.rrset_prev->next = rrset;
+		else	state->c.domain->rrsets = rrset;
+		region_recycle(state->database->region, rrset_orig,
+			sizeof(rrset_type) +
+			rrset_orig->rr_count*sizeof(rr_type*));
+#endif /* PACKED_STRUCTS */
+		memcpy(rrset->rrs + rrset->rr_count, state->c.rrs, state->c.rr_count * sizeof(rr_type*));
+		rrset->rr_count += state->c.rr_count;
+	}
+	state->records += state->c.rr_count;;
+	/* Check we have SOA */
+	if (state->c.rrs[0]->owner == state->zone->apex)
+		apex_rrset_checks(state->database, rrset, state->c.rrs[0]->owner);
+
+	state->c.domain = NULL;
+	state->c.type = -1;
+	state->c.rrset = NULL;
+#ifdef PACKED_STRUCTS
+	state->c.rrset_prev = NULL;
+#endif
+	state->c.rr_count = 0;
+}
+
 int32_t zonec_accept(
 	zone_parser_t *parser,
 	const zone_name_t *owner,
@@ -180,36 +212,53 @@ int32_t zonec_accept(
 	void *user_data)
 {
 	struct rr *rr;
-	struct rrset *rrset;
-	const struct dname *dname;
+	struct dname_buffer dname;
 	struct domain *domain;
 	struct buffer buffer;
 	int priority;
-	union rdata_atom *rdatas;
-	ssize_t rdata_count;
+	int32_t code;
+	const struct nsd_type_descriptor *descriptor;
 	struct zonec_state *state = (struct zonec_state *)user_data;
-
 	assert(state);
 
-	// emulate packet buffer to leverage rdata_wireformat_to_rdata_atoms
 	buffer_create_from(&buffer, rdata, rdlength);
 
 	priority = parser->options.secondary ? ZONE_WARNING : ZONE_ERROR;
-	// limit to IN class
+	/* limit to IN class */
 	if (class != CLASS_IN)
 		zone_log(parser, priority, "only class IN is supported");
 
-	dname = dname_make(state->rr_region, owner->octets, 1);
-	assert(dname);
-	domain = domain_table_insert(state->domains, dname);
+	if(!dname_make_buffered(&dname, (uint8_t*)owner->octets, 1)) {
+		zone_log(parser, ZONE_ERROR, "the owner cannot be converted");
+		return ZONE_BAD_PARAMETER;
+	}
+	domain = domain_table_insert(state->domains, (void*)&dname);
 	assert(domain);
-
-	rdatas = NULL;
-	rdata_count = rdata_wireformat_to_rdata_atoms(
-		state->database->region, state->domains, type, rdlength, &buffer, &rdatas);
-	// number of atoms must not exceed maximum of 65535 (all empty strings)
-	assert(rdata_count >= 0);
-	assert(rdata_count <= MAX_RDLENGTH);
+	if (domain != state->c.domain || type != state->c.type
+	||  state->c.rr_count >= (int)(sizeof(state->c.rrs) / sizeof(*state->c.rrs))){
+		zonec_commit_rrset(parser, state);
+		state->c.domain = domain;
+		state->c.type = type;
+		state->c.rrset = NULL;
+#ifdef PACKED_STRUCTS
+		state->c.rrset_prev = NULL;
+#endif
+	}
+	descriptor = nsd_type_descriptor(type);
+	code = descriptor->read_rdata(state->domains, rdlength, &buffer, &rr);
+	if(code < 0) {
+		zone_log(parser, ZONE_ERROR, "the RR rdata fields are wrong for the type, %s %s %s",
+			dname_to_string((void*)&dname,0),
+			rrtype_to_string(type),
+			read_rdata_fail_str(code));
+		if(code == TRUNCATED)
+			return ZONE_OUT_OF_MEMORY;
+		return ZONE_BAD_PARAMETER;
+	}
+	rr->owner = domain;
+	rr->type = type;
+	rr->klass = class;
+	rr->ttl = ttl;
 
 	/* we have the zone already */
 	if (type == TYPE_SOA) {
@@ -230,65 +279,33 @@ int32_t zonec_accept(
 		zone_log(parser, priority, "out of zone data: %s is outside the zone for fqdn %s",
 		         s, domain_to_string(domain));
 		if (!parser->options.secondary) {
-			region_free_all(state->rr_region);
 			return ZONE_SEMANTIC_ERROR;
 		}
 	}
+	/* With the first RR for a RRset in this position in the zone file,
+	 * find the RRset */
+	if (state->c.rr_count == 0)  {
+#ifndef PACKED_STRUCTS
+		state->c.rrset = domain_find_rrset(state->c.domain, state->zone, state->c.type);
+#else
+		state->c.rrset = domain_find_rrset_and_prev(state->c.domain, state->zone, state->c.type, &state->c.rrset_prev);
+#endif
+	}
+	if (type == TYPE_RRSIG)
+		; /* pass */
+	else if (state->c.rrset && ttl != state->c.rrset->rrs[0]->ttl) {
+		zone_log(parser, ZONE_WARNING,
+			"%s TTL %"PRIu32" does not match TTL %u of %s RRset",
+			domain_to_string(domain), ttl,
+			state->c.rrset->rrs[0]->ttl, rrtype_to_string(type));
 
-	/* Do we have this type of rrset already? */
-	rrset = domain_find_rrset(domain, state->zone, type);
-	if (!rrset) {
-		rrset = region_alloc(state->database->region, sizeof(*rrset));
-		rrset->zone = state->zone;
-		rrset->rr_count = 0;
-		rrset->rrs = region_alloc(state->database->region, sizeof(*rr));
-
-		switch (type) {
-			case TYPE_CNAME:
-				if (!domain_find_non_cname_rrset(domain, state->zone))
-					break;
-				zone_log(parser, priority, "CNAME and other data at the same name");
-				break;
-			case TYPE_RRSIG:
-			case TYPE_NXT:
-			case TYPE_SIG:
-			case TYPE_NSEC:
-			case TYPE_NSEC3:
-				break;
-			default:
-				if (!domain_find_rrset(domain, state->zone, TYPE_CNAME))
-					break;
-				zone_log(parser, priority, "CNAME and other data at the same name");
-				break;
-		}
-
-		/* Add it */
-		domain_add_rrset(domain, rrset);
-	} else {
-		struct rr *rrs;
-		if (type != TYPE_RRSIG && ttl != rrset->rrs[0].ttl) {
-			zone_log(parser, ZONE_WARNING, "%s TTL %"PRIu32" does not match TTL %u of %s RRset",
-				domain_to_string(domain), ttl, rrset->rrs[0].ttl,
-					rrtype_to_string(type));
-		}
-
-		/* Search for possible duplicates... */
-		for (int i = 0; i < rrset->rr_count; i++) {
-			if (zrdatacmp(type, rdatas, rdata_count, &rrset->rrs[i]) != 0)
-				continue;
-			/* Discard the duplicates... */
-			for (size_t j = 0; j < (size_t)rdata_count; j++) {
-				size_t size;
-				if (rdata_atom_is_domain(type, j))
-					continue;
-				size = rdata_atom_size(rdatas[j]) + sizeof(uint16_t);
-				region_recycle(state->database->region, rdatas[j].data, size);
-			}
-			region_recycle(state->database->region, rdatas, sizeof(*rdatas) * rdata_count);
-			region_free_all(state->rr_region);
-			return 0;
-		}
-
+	} else if (state->c.rr_count && ttl != state->c.rrs[0]->ttl) {
+		zone_log(parser, ZONE_WARNING,
+			"%s TTL %"PRIu32" does not match TTL %u of %s RRset",
+			domain_to_string(domain), ttl,
+			state->c.rrs[0]->ttl, rrtype_to_string(type));
+	}
+	if (state->c.rrset || state->c.rr_count) {
 		switch (type) {
 			case TYPE_CNAME:
 				zone_log(parser, priority, "multiple CNAMEs at the same name");
@@ -299,28 +316,30 @@ int32_t zonec_accept(
 			default:
 				break;
 		}
-
-		/* Add it... */
-		rrs = rrset->rrs;
-		rrset->rrs = region_alloc_array(state->database->region, rrset->rr_count + 1, sizeof(*rr));
-		memcpy(rrset->rrs, rrs, rrset->rr_count * sizeof(*rr));
-		region_recycle(state->database->region, rrs, rrset->rr_count * sizeof(*rr));
 	}
-
-	rr = &rrset->rrs[rrset->rr_count++];
-	rr->owner = domain;
-	rr->rdatas = rdatas;
-	rr->ttl = ttl;
-	rr->type = type;
-	rr->klass = class;
-	rr->rdata_count = rdata_count;
-
-	/* Check we have SOA */
-	if (rr->owner == state->zone->apex)
-		apex_rrset_checks(state->database, rrset, rr->owner);
-
-	state->records++;
-	region_free_all(state->rr_region);
+	if (state->c.rrset) {
+		/* Search for possible duplicates in existing RRset */
+		for (int i = 0; i < state->c.rrset->rr_count; i++) {
+			if (!equal_rr_rdata(descriptor, rr, state->c.rrset->rrs[i]))
+				continue;
+			/* Discard the duplicates... */
+			/* Lower the usage counter for domains in the rdata. */
+			rr_lower_usage(state->database, rr);
+			region_recycle(state->database->region, rr, sizeof(*rr) + rr->rdlength);
+			return 0;
+		}
+	}
+	/* Search for possible duplicates in already batched RRs */
+	for (int i = 0; i < state->c.rr_count; i++) {
+		if (!equal_rr_rdata(descriptor, rr, state->c.rrs[i]))
+			continue;
+		/* Discard the duplicates... */
+		/* Lower the usage counter for domains in the rdata. */
+		rr_lower_usage(state->database, rr);
+		region_recycle(state->database->region, rr, sizeof(*rr) + rr->rdlength);
+		return 0;
+	}
+	state->c.rrs[state->c.rr_count++] = rr;
 	return 0;
 }
 
@@ -416,12 +435,17 @@ zonec_read(
 
 	state.database = database;
 	state.domains = domains;
-	state.rr_region = region_create(xalloc, free);
 	state.zone = zone;
-	state.domain = NULL;
 	state.errors = 0;
 	state.records = 0;
 
+	state.c.domain = NULL;
+	state.c.type = -1;
+	state.c.rrset = NULL;
+#ifdef PACKED_STRUCTS
+	state.c.rrset_prev = NULL;
+#endif
+	state.c.rr_count = 0;
 	origin = domain_dname(zone->apex);
 	memset(&options, 0, sizeof(options));
 	options.origin.octets = dname_name(origin);
@@ -436,9 +460,19 @@ zonec_read(
 
 	/* Parse and process all RRs.  */
 	if (zone_parse(&parser, &options, &buffers, zonefile, &state) != 0) {
-		region_destroy(state.rr_region);
+		/* With all socked up RRs,
+		 * lower the usage counter for domains in the rdata.
+		 */
+		for (int i = 0; i < state.c.rr_count; i++) {
+			/* Lower the usage counter for domains in the rdata. */
+			rr_lower_usage(database, state.c.rrs[i]);
+			region_recycle( database->region, state.c.rrs[i]
+			              , sizeof(*state.c.rrs[i])
+			              + state.c.rrs[i]->rdlength);
+		}
 		return state.errors;
 	}
+	zonec_commit_rrset(&parser, &state);
 
 	/* Check if zone file contained a correct SOA record */
 	if (!zone) {
@@ -447,23 +481,22 @@ zonec_read(
 	} else if (!zone->soa_rrset || zone->soa_rrset->rr_count == 0) {
 		log_msg(LOG_ERR, "zone configured as '%s' has no SOA record", name);
 		state.errors++;
-	} else if (dname_compare(domain_dname(zone->soa_rrset->rrs[0].owner), origin) != 0) {
+	} else if (dname_compare(domain_dname(zone->soa_rrset->rrs[0]->owner), origin) != 0) {
 		log_msg(LOG_ERR, "zone configured as '%s', but SOA has owner '%s'",
-		        name, domain_to_string(zone->soa_rrset->rrs[0].owner));
+		        name, domain_to_string(zone->soa_rrset->rrs[0]->owner));
 		state.errors++;
 	}
 
 	if(!zone_is_slave(zone->opts) && !check_dname(zone))
 		state.errors++;
 
-	region_destroy(state.rr_region);
 	return state.errors;
 }
 
 void
 apex_rrset_checks(namedb_type* db, rrset_type* rrset, domain_type* domain)
 {
-	uint32_t soa_minimum;
+	uint32_t soa_minimum = 0;
 	unsigned i;
 	zone_type* zone = rrset->zone;
 	assert(domain == zone->apex);
@@ -474,26 +507,34 @@ apex_rrset_checks(namedb_type* db, rrset_type* rrset, domain_type* domain)
 		/* BUG #103 add another soa with a tweaked ttl */
 		if(zone->soa_nx_rrset == 0) {
 			zone->soa_nx_rrset = region_alloc(db->region,
-				sizeof(rrset_type));
+				sizeof(rrset_type)
+#ifdef PACKED_STRUCTS
+				+ sizeof(rr_type*)
+#endif
+				);
 			zone->soa_nx_rrset->rr_count = 1;
 			zone->soa_nx_rrset->next = 0;
 			zone->soa_nx_rrset->zone = zone;
+#ifndef PACKED_STRUCTS
 			zone->soa_nx_rrset->rrs = region_alloc(db->region,
-				sizeof(rr_type));
+				sizeof(rr_type*));
+#endif
+			zone->soa_nx_rrset->rrs[0] = region_alloc(db->region,
+				sizeof(rr_type)+rrset->rrs[0]->rdlength);
 		}
-		memcpy(zone->soa_nx_rrset->rrs, rrset->rrs, sizeof(rr_type));
+		memcpy(zone->soa_nx_rrset->rrs[0], rrset->rrs[0],
+			sizeof(rr_type)+rrset->rrs[0]->rdlength);
 
 		/* check the ttl and MINIMUM value and set accordingly */
-		memcpy(&soa_minimum, rdata_atom_data(rrset->rrs->rdatas[6]),
-				rdata_atom_size(rrset->rrs->rdatas[6]));
-		if (rrset->rrs->ttl > ntohl(soa_minimum)) {
-			zone->soa_nx_rrset->rrs[0].ttl = ntohl(soa_minimum);
+		retrieve_soa_rdata_minttl(rrset->rrs[0], &soa_minimum);
+		if (rrset->rrs[0]->ttl > soa_minimum) {
+			zone->soa_nx_rrset->rrs[0]->ttl = soa_minimum;
 		}
 	} else if (rrset_rrtype(rrset) == TYPE_NS) {
 		zone->ns_rrset = rrset;
 	} else if (rrset_rrtype(rrset) == TYPE_RRSIG) {
 		for (i = 0; i < rrset->rr_count; ++i) {
-			if(rr_rrsig_type_covered(&rrset->rrs[i])==TYPE_DNSKEY){
+			if(rr_rrsig_type_covered(rrset->rrs[i])==TYPE_DNSKEY){
 				zone->is_secure = 1;
 				break;
 			}
